{% extends base_template|default:"django_fido/base.html" %}
{% load static %}

{% block extrahead %}
    {{ block.super }}
    {{ form.media }}
{% endblock %}

{% block content %}
    <form method="post" id="django-fido-form" data-mode="{{ view.fido2_request_type }}" data-url="{{ view.fido2_request_url }}" data-autosubmit-off>
        <ul id="django-fido-errors">
        </ul>
        {% csrf_token %}
        {{ form }}
        <input type="submit" id="submit-button" value="Register">
    </form>
{% endblock %}
